"""
make distrodb from apt-rpm pkglist.
"""
import os, fnmatch
import sys
import rpm
import re
from .decompress import DecompressedTmpFile, supported_suffixes
import DistroDbMaker.XMLParser

arg_pkglist_pattern=re.compile('^.*(?:hd|pkg|src)list(?:.*)$')
arg_dir_repodata_pattern=re.compile('^(?:|.*/)repodata/?$')
arg_yum_sqlite_pattern=re.compile('^.*primary\.sqlite.*$')
arg_yum_primary_xml_pattern=re.compile('^.*primary\.xml.*$')

def find(pattern, path):
    result = []
    for root, dirs, files in os.walk(path):
        for name in files:
            if fnmatch.fnmatch(name, pattern):
                result.append(os.path.join(root, name))
    return result

def find_first_compressed(pattern, path):
    result = []
    for suffix in supported_suffixes():
        for root, dirs, files in os.walk(path):
            for name in files:
                if fnmatch.fnmatch(name, pattern + suffix):
                    result.append(os.path.join(root, name))
        if result:
            return result
    return result

def readBoolMap(filename):
    """ Read into a list. """
    data={}
    f = open(filename, "ro")
    l=f.readline()
    while len(l)>1:
        data[l.strip()]=True
        l=f.readline()
    f.close()
    return data

rpm5=False
def readRpmHeader(filename):
    """ Read an rpm header. """
    fd = os.open(filename, os.O_RDONLY)
    isSource = False
    if hasattr(rpm, '__version__'):
        # rpm > 4.04, expect rpm 4.13
        ts = rpm.TransactionSet()
        header = ts.hdrFromFdno(fd)
        isSource = header[rpm.RPMTAG_SOURCEPACKAGE]
    else:
        # rpm 4.04
        (header, isSource) = rpm.headerFromPackage(fd)
    os.close(fd)
    if rpm5 or header['RPMVERSION'][0]=='5':
        #print "rpm5: %s: header['SOURCERPM'] = %s" % (filename,header['SOURCERPM'])
        isSource = not header['SOURCERPM']
    return (header, isSource)

class DistroDBCLI(object):

    def __init__(self, verbose=0):
        self.verbose=verbose
        self.debug=0
        self.opt_handlers={}
        self.longopt=[]
        self.set_default_opt_handlers()

    def set_default_opt_handlers(self):
        # --- built-in ---
        self.add_longopt("repo=")
        self.add_longopt("branch=")
        self.add_longopt("component=")

    def usage(self):
        print """create distrodb
usage:
-b branch
-r repo
-h		this help
-v num		debug level
"""

    def set_opt_handler(self,option,code):
        self.opt_handlers[option]=code

    def add_longopt(self,longopt):
        self.longopt.append(longopt)

    def get_commandline_opt(self):
        import getopt
        self.repo=None
        self.branch='default'
        self.component='release'
        try:
            opts, args = getopt.getopt(sys.argv[1:], "Db:hr:v", self.longopt)
            self.cli_args=args
        except getopt.GetoptError:
            # print help information and exit:
            self.usage()
            sys.exit(2)
        for o, a in opts:
            if o in self.opt_handlers:
                self.opt_handlers[o](self, o, a)
            elif o == "-D":
                self.debug += 1
            elif o == "-v":
                self.verbose += 1
            elif o in ("-b", "--branch"):
                self.branch=a;
            elif o in ("-r", "--repo"):
                self.repo=a;
            elif o in ("-c", "--component"):
                self.component=a;
            elif o in ("-h", "--help"):
                self.usage()
                sys.exit()
        if self.repo == None:
            print "Error: --repo option is mandatory!"
            self.usage()
            sys.exit()

class RepoReader(DistroDBCLI):

    def __init__(self, verbose=False):
        super(RepoReader,self).__init__(verbose)
        self.srpm_filter=None
        self.rpmheader_handler=None
        self.cli_ignore_srpms_in_binary=False
        self.process_repodata_primary_xml=DistroDbMaker.XMLParser.process_repodata_primary_xml_parse
        # not much time difference on big sets, but saves memeory
        #self.process_repodata_primary_xml=DistroDbMaker.XMLParser.process_repodata_primary_xml_iter3
        self.process_repodata_filelists_xml=DistroDbMaker.XMLParser.process_repodata_filelists_xml
        self.rawrpm=[]
        self.primary_xml=[]
        self.filelists_xml=[]
        
    def usage(self):
        super(RepoReader,self).usage()
        print '--name-filter <file> TODO - not implemented yet'
        print '--rpm5'
        print '--ignore-srpms-in-binary'

    def set_default_opt_handlers(self):
        super(RepoReader,self).set_default_opt_handlers()
        self.add_longopt('rpm5')
        self.set_opt_handler('--rpm5', lambda reader,o,a: reader.set_rpm5(True))
        self.add_longopt("name-file=")
        self.set_opt_handler('--name-file', lambda reader,o,a: reader.set_name_filter(a))
        self.add_longopt("name=")
        self.set_opt_handler('--name', lambda reader,o,a: reader.rpmheader_handler.name_filter.append(a))
        self.add_longopt('exclude-nosrc')
        self.set_opt_handler('--exclude-nosrc', lambda reader,o,a: reader.rpmheader_handler.set_exclude_nosrc(True))
        self.add_longopt('ignore-srpms-in-binary')
        self.set_opt_handler('--ignore-srpms-in-binary', lambda reader,o,a: reader.set_ignore_srpms_in_binary(True))

    def set_name_filter(self, namelistfile):
        if namelistfile:
            self.rpmheader_handler.name_filter=readBoolMap(namelistfile)

    def set_rpm5(self, rpm5flag):
        rpm5=rpm5flag
    def set_ignore_srpms_in_binary(self, flag):
        self.cli_ignore_srpms_in_binary=flag

    def process_args(self, args):
        verbose=self.verbose
        rpmheader_handler=self.rpmheader_handler
        rpmheader_handler.set_verbose(verbose)
        rpmheader_handler.set_debug(self.debug)
        rawrpm=self.rawrpm
        primary_xml=self.primary_xml
        filelists_xml=self.filelists_xml
        for arg in args:
            if arg_pkglist_pattern.match(arg):
                print "processing pkglist %s..." % arg
                self.process_rpm_pkglist(arg, self.rpmheader_handler)
            elif arg[-8:]=='.src.rpm':
                rawrpm.append(arg);
            elif os.path.isdir(arg):
                if arg_dir_repodata_pattern.match(arg):
                    #print "processing repodata at %s..." % arg
                    for xml in find_first_compressed('*primary.xml', arg):
                        primary_xml.append(xml)
                    for xml in find_first_compressed('*filelists.xml', arg):
                        filelists_xml.append(xml)
                else:
                    print "processing rpms in dir %s..." % arg
                    for rpm in find('*.rpm', arg):
                        rawrpm.append(rpm)
            elif arg_yum_primary_xml_pattern.match(arg):
                primary_xml.append(arg)
            elif arg_yum_sqlite_pattern.match(arg):
                print "TODO processing yum sqlite db %s..." % arg
                pass
        if len(primary_xml):
            for xml in primary_xml:
                print "processing primary.xml repodata at %s..." % xml
                self.process_repodata_primary_xml(xml, rpmheader_handler)
        if len(filelists_xml) and self.process_repodata_filelists_xml:
            for xml in filelists_xml:
                print "processing filelists.xml repodata at %s..." % xml
                self.process_repodata_filelists_xml(xml, rpmheader_handler)
        if len(rawrpm):
            print "processing rpms ..."
            self.process_raw_rpms(rawrpm, rpmheader_handler)

    def process_raw_rpms(self,rpmlist,rpmheader_handler):
        verbose=self.verbose
        for rpmfile in rpmlist:
            if verbose:
                print "processing rpm %s..." % rpmfile
            self.process_raw_rpm(rpmfile, rpmheader_handler)

    def process_raw_rpm(self, rpmfile, rpmheader_handler):
        (header, isSource)=readRpmHeader(rpmfile)
        rpmheader_handler.process_rpm_header(header, isSource, rpmfile)
                
    def process_rpm_pkglist(self, arg, rpmheader_handler):
        tmpfile=decompress.DecompressedTmpFile(arg)
        rpmheaders = rpm.readHeaderListFromFile(tmpfile.get_tmpfile_path())
        tmpfile.cleanup_tmpfile()
        for rpmheader in rpmheaders:
            rpmheader_handler.process_rpm_header(rpmheader)

    def run(self):
        self.get_commandline_opt()
        self.process_args(self.cli_args)
        distrodb=Writer(self.repo,self.branch,self.component)
        self.rpmheader_handler.write_distrodb(distrodb)


class RPM2DistroDB(object):

    def __init__(self, verbose=0):
        #super(RPM2DistroDB,self).__init__(verbose)
        self.verbose=verbose
        self.debug=0
        self.name_filter=[]
        self.exclude_nosrc=[]

    def set_exclude_nosrc(self,val):
        self.exclude_nosrc=val

    def set_verbose(self,verbose):
        self.verbose=verbose

    def set_debug(self,debug):
        #print "RPM2DistroDB %s: debug is set to %s" % (self,debug)
        self.debug=debug

    def process_rpm_header(self,rpmheader,isSource=False,filepath=None):
        pass

    def process_filelist_xml(self,filelist,name,arch,isSource=False):
        if not name:
            raise ValueError('Oops! process_filelist_xml: got no name')
        if self.arch_check(arch):
            self.process_filelist(filelist, self.mangle_name(name))

    def process_filelist(self,filelist,name):
        pass

    def arch_check(self,arch):
        return True

    def mangle_name(self,name):
        return name

    def write_distrodb(self,distrodb):
        pass

class DispatcherRPM2DistroDB(RPM2DistroDB):

    def __init__(self, srchandler=None, binhandler=None, verbose=False):
        super(DispatcherRPM2DistroDB,self).__init__(verbose)
        self.verbose=verbose
        self.srchandler=srchandler
        self.binhandler=binhandler
        #self.srpm_filter=None
        
    def set_exclude_nosrc(self,val):
        self.srchandler.set_exclude_nosrc(val)

    def set_verbose(self,verbose):
        self.verbose=verbose
        self.srchandler.verbose=verbose
        self.binhandler.verbose=verbose

    def set_debug(self,debug):
        self.debug=debug
        self.srchandler.debug=debug
        self.binhandler.debug=debug

    def process_rpm_header(self,rpmheader,isSource=False,filepath=None):
        if isSource:
            srchandler=self.srchandler
            if srchandler:
                srchandler.process_rpm_header(rpmheader,isSource,filepath)
            elif self.die_on_src_bin_mismatch:
                raise ValueError("Oops! header for %s is not from a source package" % rpmheader['NAME'])
        else:
            binhandler=self.binhandler
            if binhandler:
                binhandler.process_rpm_header(rpmheader,isSource,filepath)
            elif self.die_on_src_bin_mismatch:
                raise ValueError("Oops! header for %s is not from a binary package" % rpmheader['NAME'])

    def process_filelist_xml(self,filelist,name,arch,isSource=False):
        if isSource:
            return
        binhandler=self.binhandler
        if not binhandler:
            raise ValueError("Oops! no binhandler for a filelist xml - name=%s" % name)
        if self.name_filter and not name in self.name_filter:
            return
        if name and binhandler.arch_check(arch):
            binhandler.process_filelist(filelist, binhandler.mangle_name(name))

    def write_distrodb(self,distrodb):
        if self.srchandler:
            self.srchandler.write_distrodb(distrodb)
        if self.binhandler:
            self.binhandler.write_distrodb(distrodb)

class Writer(object):
    """class for pattern db."""
    def __init__(self, repo='generic', branch='default',component='release'):
        self.repo=repo
        self.branch = branch
        self.component=component
        self.root=os.environ['HOME']+"/.cache/distrodb/"+self.repo+'/'+self.branch
        if not os.path.isdir(self.root):
            os.makedirs(self.root)
        self.main_db_version=11
        self.main_db_compat=5
        self.raw_db_version=4
        self.raw_db_compat=0

    def write_version(self):
        self._write_verfile(self.root+'/'+'.version',self.main_db_version)
        self._write_verfile(self.root+'/'+'.compat',self.main_db_compat)
        self._write_verfile(self.root+'/'+'version.raw',self.raw_db_version)
        self._write_verfile(self.root+'/'+'compat.raw',self.raw_db_compat)

    def _write_verfile(self,verfile,ver):
        f=open(verfile,'w')
        f.write("%s\n" % ver)
        f.close()

    def writemap(self,dir,db):
        path=self.root+'/'+dir
        if not os.path.isdir(path):
            os.makedirs(path)
        mfile=path+'/'+self.component+".txt"
        #print "mfile=%s" % mfile
        f=open(mfile,'w')
        try:
            db.write_to_f(f)
        except (UnicodeDecodeError,UnicodeEncodeError):
            print "Writer: unicode error in %s" % path
            sys.exit(1)
        f.close()


"""
.output $distrodbdir/perl/${component}.txt
select providename, name from rpm_provides join rpm on rpm_provides.pkgid=rpm.pkgid where providename glob 'perl(*)';

.schema files
CREATE TABLE files (  name TEXT,  type TEXT,  pkgKey INTEGER);
CREATE INDEX filenames ON files (name);
CREATE INDEX pkgfiles ON files (pkgKey);
sqlite> .schema packages
CREATE TABLE packages (  pkgKey INTEGER PRIMARY KEY,  pkgId TEXT,  name TEXT,  arch TEXT,  version TEXT,  epoch TEXT,  release TEXT,  summary TEXT,  description TEXT,  url TEXT,  time_file INTEGER,  time_build INTEGER,  rpm_license TEXT,  rpm_vendor TEXT,  rpm_group TEXT,  rpm_buildhost TEXT,  rpm_sourcerpm TEXT,  rpm_header_start INTEGER,  rpm_header_end INTEGER,  rpm_packager TEXT,  size_package INTEGER,  size_installed INTEGER,  size_archive INTEGER,  location_href TEXT,  location_base TEXT,  checksum_type TEXT);
sqlite> select * from files limit 1;
/etc/ovirt-engine/engine-manage-domains|dir|6

sqlite> .schema provides
CREATE TABLE provides (  name TEXT,  flags TEXT,  epoch TEXT,  version TEXT,  release TEXT,  pkgKey INTEGER );
CREATE INDEX pkgprovides on provides (pkgKey);
CREATE INDEX providesname ON provides (name);

"""
